require 'pastel'

module Misfortune
  module Render
    class Terminal < Redcarpet::Render::StripDown
      BLOCK_MAX_WITH = 80
      BLOCK_DELIMITER = '~'.freeze
      LIST_INDENTATION = '  '.freeze
      LIST_ORDERED_ITEM_PREFIX = '<ENUM>'.freeze
      LIST_ORDERED_ITEM_CHAR = '.'.freeze
      LIST_UNORDERED_ITEM_CHAR = '-'.freeze

      attr_reader :pastel

      def initialize
        @pastel = Pastel.new(enabled: $stdout.tty?, eachline: new_line)

        super
      end

      def header(text, _header_level)
        text = pastel.underline(text)
        surround_with_line_breaks(text, 0, 2)
      end

      def block_code(code, language)
        text = ''
        text << labeled_hrule(language)
        text << indent(code(code))
        text << hrule

        surround_with_line_breaks(text, 1, 2)
      end

      def block_quote(quote)
        text = pastel.cyan(quote)
        surround_with_line_breaks(text, 1, 2)
      end

      def paragraph(text)
        surround_with_line_breaks(text, 0, 2)
      end

      def list(contents, list_type)
        contents = enumerate_list(contents) if list_type == :ordered
        contents = indent(contents)

        surround_with_line_breaks(contents, 0, 1)
      end

      def list_item(text, list_type)
        prefix =
          case list_type
          when :ordered
            LIST_ORDERED_ITEM_PREFIX
          when :unordered
            LIST_UNORDERED_ITEM_CHAR
          end

        "#{prefix} #{text}"
      end

      def hrule
        labeled_hrule('')
      end

      def emphasis(text)
        pastel.underline(text)
      end

      def double_emphasis(text)
        pastel.bold(text)
      end

      def triple_emphasis(text)
        emphasis(double_emphasis(text))
      end

      def strikethrough(text)
        pastel.strikethrough(text)
      end

      def link(link, _title, content)
        "#{content} (#{pastel.blue(link)})"
      end

      def autolink(link, _link_type)
        pastel.blue(link)
      end

      def codespan(code)
        code(code)
      end

      protected

      def surround_with_line_breaks(text, prepend_count, append_count)
        text = clear_surrounding_line_breaks(text)
        (new_line * prepend_count) << text << (new_line * append_count)
      end

      def clear_surrounding_line_breaks(text)
        text.sub(/\A[\r\n]+/, '').sub(/[\r\n]+\z/, '')
      end

      def indent(text)
        text.strip
          .gsub(new_line, "#{new_line}#{LIST_INDENTATION}")
          .prepend(LIST_INDENTATION)
      end

      def enumerate_list(text)
        text.gsub(LIST_ORDERED_ITEM_PREFIX).with_index do |_, i|
          "#{i + 1}#{LIST_ORDERED_ITEM_CHAR}"
        end
      end

      def new_line
        "\n"
      end

      def code(text)
        pastel.green(text.chomp)
      end

      def labeled_hrule(text)
        text = (text || '').to_s.upcase
        text = text.center(10, ' ') unless text.empty?
        text = text.center(BLOCK_MAX_WITH, BLOCK_DELIMITER)

        surround_with_line_breaks(text, 1, 1)
      end
    end
  end
end
